/* 
 * Copyright (C) 2006 Registro.br. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 1. Redistribution of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY REGISTRO.BR ``AS IS AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIE OF FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL REGISTRO.BR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
/* $Id: DomParser.H 372 2006-02-09 19:21:57Z eduardo $ */
/** @file DomParser.H
 *  @brief EPP DomParser Class
 */

#ifndef __DOMPARSER_H__
#define __DOMPARSER_H__

#include <memory>

#include <xercesc/dom/DOM.hpp>

#include "libepp_nicbr.H"

#include "Response.H"
#include "Greeting.H"
#include "ContactCheckRsp.H"
#include "ContactCreateRsp.H"
#include "ContactInfoRsp.H"
#include "DomainCheckRsp.H"
#include "DomainCreateRsp.H"
#include "DomainInfoRsp.H"
#include "DomainRenewRsp.H"
#include "PollRsp.H"
#include "DomErrorHandler.H"

using std::auto_ptr;

XERCES_CPP_NAMESPACE_USE

LIBEPP_NICBR_NS_BEGIN

/// EPP DomParser Class
class DomParser
{
public:
  /// Constructor
  DomParser();

  /// Destructor
  ~DomParser();

  /// Enable Validation
  /**
     @param schemas_dir Path to the schemas directory
   */
  void enable_validation(const string& schemas_dir = SCHEMASDIR);

  /// Disable Validation
  void disable_validation();

  /// Parse method to be used by the Unit Test to validate commands
  /**
     @param xml_payload Command (in XML format) to be validated
   */
  void parse_command(const string &xml_payload);

  /// Parse the response to a login command
  /**
     @param xml_payload XML to be parsed
     @return Auto pointer to a response
   */
  auto_ptr<Response> parse_login_rsp(const string &xml_payload);

  /// Parse the response to a logout command
  /**
     @param xml_payload XML to be parsed
     @return Auto pointer to a response
   */
  auto_ptr<Response> parse_logout_rsp(const string &xml_payload);

  /// Parse the response to a connect or hello command
  /**
     @param xml_payload XML to be parsed
     @return Auto pointer to a greeting
   */
  auto_ptr<Greeting> parse_greeting(const string &xml_payload);

  /// Parse the response to a contact check command
  /**
     @param xml_payload XML to be parsed
     @return Auto pointer to a contact check response
   */
  auto_ptr<ContactCheckRsp> parse_contact_check_rsp(const string &xml_payload);

  /// Parse the response to a contact create command
  /**
     @param xml_payload XML to be parsed
     @return Auto pointer to a contact create response
   */
  auto_ptr<ContactCreateRsp> parse_contact_create_rsp(
                                                    const string &xml_payload);

  /// Parse the response to a contact info command
  /**
     @param xml_payload XML to be parsed
     @return Auto pointer to a contact info response
   */
  auto_ptr<ContactInfoRsp> parse_contact_info_rsp(const string &xml_payload);

  /// Parse the response to a contact update command
  /**
     @param xml_payload XML to be parsed
     @return Auto pointer to a contact update response
   */
  auto_ptr<Response> parse_contact_update_rsp(const string &xml_payload);

  /// Parse the response to a domain check command
  /**
     @param xml_payload XML to be parsed
     @return Auto pointer to a domain check response
   */
  auto_ptr<DomainCheckRsp> parse_domain_check_rsp(const string &xml_payload);

  /// Parse the response to a domain create command
  /**
     @param xml_payload XML to be parsed
     @return Auto pointer to a domain create response
   */
  auto_ptr<DomainCreateRsp> parse_domain_create_rsp(const string &xml_payload);

  /// Parse the response to a domain info command
  /**
     @param xml_payload XML to be parsed
     @return Auto pointer to a domain info response
   */
  auto_ptr<DomainInfoRsp> parse_domain_info_rsp(const string &xml_payload);

  /// Parse the response to a domain renew command
  /**
     @param xml_payload XML to be parsed
     @return Auto pointer to a domain renew response
   */
  auto_ptr<DomainRenewRsp> parse_domain_renew_rsp(const string &xml_payload);

  /// Parse the response to a domain update command
  /**
     @param xml_payload XML to be parsed
     @return Auto pointer to a domain update response
   */
  auto_ptr<Response> parse_domain_update_rsp(const string &xml_payload);

  /// Parse the response to a poll command
  /**
     @param xml_payload XML to be parsed
     @return Auto pointer to a poll response
  */
  auto_ptr<PollRsp> parse_poll_rsp(const string &xml_payload);

  /// Transcodes a XMLCh to string
  /**
     @param to_transcode String to be transcoded
     @return Transcoded string
   */
  static string str_transcode(const XMLCh *const to_transcode);

protected:
  /// Do the actual parsing of the XML document
  /**
     @param xml_payload XML to be parsed
     @return Pointer to the root of the document tree
   */
  DOMNode *parse(const string &xml_payload);

  /// Parse the response
  /**
     @param xml_payload XML to be parsed
     @return Auto pointer to a response
   */
  auto_ptr<Response> parse_response(const string &xml_payload);

  /// Fills a Response object 
  /**
     @param n Pointer to a node of the document tree
     @param response Pointer to a Response object
   */
  void fill_response(DOMNode *n, Response *response);

  /// Fills a Greeting object
  /**
     @param n Pointer to a node of the document tree
     @param greeting Pointer to a greeting object
   */
  void fill_greeting(DOMNode *n, Greeting *greeting);

  /// Fills a Contact check response object
  /**
     @param n Pointer to a node of the document tree
     @param c_check_rsp Pointer to a contact check response
   */
  void fill_contact_check_rsp(DOMNode *n, ContactCheckRsp *c_check_rsp);

  /// Fills a Contact create response object
  /**
     @param n Pointer to a node of the document tree
     @param c_create_rsp Pointer to a contact create response
   */
  void fill_contact_create_rsp(DOMNode *n, ContactCreateRsp *c_create_rsp);

  /// Fills a Contact info response object
  /**
     @param n Pointer to a node of the document tree
     @param c_info_rsp Pointer to a contact info response
   */
  void fill_contact_info_rsp(DOMNode *n, ContactInfoRsp *c_info_rsp);

  /// Fills a Domain check response object
  /**
     @param n Pointer to a node of the document tree
     @param d_check_rsp Pointer to a domain check response
   */
  void fill_domain_check_rsp(DOMNode *n, DomainCheckRsp *d_check_rsp);

  /// Fills a Domain create response object
  /**
     @param n Pointer to a node of the document tree
     @param d_create_rsp Pointer to a domain create response
   */
  void fill_domain_create_rsp(DOMNode *n, DomainCreateRsp *d_create_rsp);

  /// Fills a Domain info response object
  /**
     @param n Pointer to a node of the document tree
     @param d_info_rsp Pointer to a domain info response
   */
  void fill_domain_info_rsp(DOMNode *n, DomainInfoRsp *d_info_rsp);

  /// Fills a Domain renew response object
  /**
     @param n Pointer to a node of the document tree
     @param d_renew_rsp Pointer to a domain renew response
  */
  void fill_domain_renew_rsp(DOMNode *n, DomainRenewRsp *d_renew_rsp);

  /// Fills a Poll response object
  /**
     @param n Pointer to a node of the document tree
     @param poll_rsp Pointer to a poll response
  */
  void fill_poll_rsp(DOMNode *n, PollRsp *poll_rsp);

  /// Fills a Postal Info struct
  /**
     @param n Pointer to a node of the document tree
     @param postal_info Pointer to a postal info struct
  */
  void fill_postal_info(DOMNode *n, struct CommonData::PostalInfo *postal_info);

  /// Object for parsing XML documents
  DOMBuilder *_builder;

  /// DOM error handler
  DomErrorHandler _error_handler;
};

LIBEPP_NICBR_NS_END
#endif //__DOMPARSER_H__
