/* 
 * Copyright (C) 2006 Registro.br. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 1. Redistribution of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY REGISTRO.BR ``AS IS AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIE OF FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL REGISTRO.BR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
/* $Id: DomainUpdateCmd.H 352 2006-02-09 16:45:19Z milena $ */
/** @file  DomainUpdateCmd.H
 *  @brief EPP DomainUpdateCmd Class
 */

#ifndef __DOMAIN_UPDATE_CMD_H__
#define __DOMAIN_UPDATE_CMD_H__

#include <string>
#include <set>
#include <map>

#include "libepp_nicbr.H"

#include "Command.H"
#include "CommonData.H" // struct NameServer

using std::string;
using std::set;
using std::map;
using std::less;

LIBEPP_NICBR_NS_BEGIN

/// EPP DomainUpdateCmd Class
class DomainUpdateCmd : public Command
{
public:
  
  struct Status {
    string s;
    string lang;
    string msg;
    bool operator<(const Status &st) const {
      return s < st.s;
    }
  };

  /// Sets domain name
  /**
     @param name   fully qualified domain name
  */
  void set_name(string name) { _name = name; }

  /// Returns domain name
  /**
     @return fully qualified domain name
  */
  string get_name() { return _name; }

  /// Inserts a nameserver_add to the set of nameservers
  /**
     @param nameserver_add   fully qualified domain name
  */
  void insert_nameserver_add(const struct NameServer &nameserver_add)
  {
    _nameserver_add.insert(nameserver_add);
  }
  
  /// Returns set of nameserver_add
  /**
     @return set of nameserver_add associated with domain object
  */
  set<struct NameServer> get_nameserver_add() { return _nameserver_add; }
  
  /// Inserts a nameserver_rem to the set of nameservers
  /**
     @param nameserver_rem   fully qualified domain name
   */
  void insert_nameserver_rem(const struct NameServer &nameserver_rem)
  {
    _nameserver_rem.insert(nameserver_rem);
  }
  
  /// Returns set of nameserver_rem
  /**
     @return set of nameserver_rem associated with domain object
  */
  set<struct NameServer> get_nameserver_rem() { return _nameserver_rem; }

  /// Inserts a contact_add in the map of contacts
  /**
     @param type_add             contact type
     @param identification_add   contact identification
  */
  void insert_contact_add(string type_add, string identification_add)
  {
    _contact_add[type_add] = identification_add;
  }
  
  /// Returns map of contacts
  /**
     @return map of contacts
  */
  map< string, string, less<string> > get_contact_add() 
  { 
    return _contact_add; 
  }
  
  /// Inserts a contact_rem in the map of contacts
  /**
     @param type_rem             contact type
     @param identification_rem   contact identification
  */
  void insert_contact_rem(string type_rem, string identification_rem)
  {
    _contact_rem[type_rem] = identification_rem;
  }
  
  /// Returns map of contacts
  /**
     @return map of contacts
  */
  map< string, string, less<string> > get_contact_rem() 
  { 
    return _contact_rem; 
  }
  
  /// Changes a status_add
  /**
     @param status_add status to be changed
  */
  void insert_status_add(const struct Status &status_add) 
  { 
    _status_add.insert(status_add);
  }
  
  /// Returns set of domain status_add
  /**
     @return set of domain status_add
  */
  set<struct Status> get_status_add() { return _status_add; }

  /// Changes a status_rem in _status_rem
  /**
     @param status_rem   the status_rem to be changed
  */
  void insert_status_rem(const struct Status &status_rem) 
  { 
    _status_rem.insert(status_rem); 
  }
  
  /// Returns set of domain status_rem
  /**
     @return set of domain status_rem
  */
  set<struct Status> get_status_rem() { return _status_rem; }
  
  // Sets registrant
  /**
     @param registrant The domain registrant
   */
  void set_registrant(const string &registrant) { _registrant = registrant; };

  // Returns registrant
  /**
     @return registrant The domain registrant
   */
  string get_registrant() { return _registrant; };

  /// Sets authorization information
  /**
     @param authInfo   domain authorization information
  */
  void set_authInfo(const AuthInfo &authInfo)
  {
    _authInfo = authInfo;
  }
  
  /// Returns authorization information
  /**
     @return authorization information
  */
  AuthInfo get_authInfo() { return _authInfo; }

  /// Reset object attributes
  void reset()
  {
    _name = "";
    _nameserver_add.erase(_nameserver_add.begin(), _nameserver_add.end());
    _nameserver_rem.erase(_nameserver_rem.begin(), _nameserver_rem.end());
    _contact_add.erase(_contact_add.begin(), _contact_add.end());
    _contact_rem.erase(_contact_rem.begin(), _contact_rem.end());
    _status_add.erase(_status_add.begin(), _status_add.end());
    _status_rem.erase(_status_rem.begin(), _status_rem.end());
    _registrant = "";
    _authInfo.reset();
  }
  
protected:
  /// fully qualified domain name
  string _name;
  
  /// name servers to be added to the domain object
  set<struct NameServer> _nameserver_add;
  
  /// name servers to be removed from the domain object
  set<struct NameServer> _nameserver_rem;
  
  /// contacts to be added to the domain object
  map< string, string, less<string> > _contact_add;

  /// contacts to be removed from the domain object
  map< string, string, less<string> > _contact_rem;
  
  /// set of status to be added to the domain object
  set<struct Status> _status_add;

  /// set of status to be removed from the domain object
  set<struct Status> _status_rem;

  /// registrant
  string _registrant;

  /// authorization information
  AuthInfo _authInfo;
};

LIBEPP_NICBR_NS_END
#endif //__DOMAIN_UPDATE_CMD_H__
