/* 
 * Copyright (C) 2006 Registro.br. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 1. Redistribution of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY REGISTRO.BR ``AS IS AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIE OF FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL REGISTRO.BR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
/* $Id: TransportTLS.H 415 2006-02-13 14:38:11Z cesar $ */
/** @file TransportTLS.H
 *  @brief TransportTLS class
 */

#ifndef __TRANSPORTTLS_H__
#define __TRANSPORTTLS_H__

#include <openssl/bio.h>
#include <openssl/ssl.h>

#include "libepp_nicbr.H"

#include "StrUtil.H"

LIBEPP_NICBR_NS_BEGIN

/// TransportTLS class
class TransportTLS
{
public:
  /// Default constructor
  TransportTLS(const string &server = "localhost", const int &port = 3121);

  /// Destructor
  ~TransportTLS();

  /// Return the server address
  /**
     @return server name/address 
  */
  const string get_server();

  /// Return the server port number
  /**
     @return server port number
  */
  const int get_port();

  /// Set the server name/address
  /**
     @param server name/address 
  */
  void set_server(const string &server);

  /// Set connection port
  /**
     @param port server port number
  */
  void set_port(const int &port);

  /// Open the Connection
  /**
     @param client_cert_file Client certificate file
     @param root_ca_file Root certificate file
     @param pem_passphrase Optional passphrase for an encrypted private key
  */
  void connect(const string &client_cert_file, const string &root_ca_file,
	       const string &pem_passphrase = "");
  
  /// Read payload from the open connection
  /**
     @param xml_payload buffer to be read
     @param timeout timeout in seconds
  */
  void read(string &xml_payload, const int &timeout = TIMEOUT);

  /// Write payload to the open connection
  /**
     @param xml_payload XML Payload
     @param timeout timeout in seconds
  */
  void write(const string &xml_payload, const int &timeout = TIMEOUT);
    
  /// Return the OpenSSL error message
  /**
     @return OpenSSL error message
  */
  string get_openssl_msg();

private:
  /// Seed OpenSSL pseudo random number generator
  void seed_prng();

  /// Callback method used by OpenSSL to collect passphrases. 
  /**
     @param buf buffer that the passphrase should be copied into
     @param size size of buf in bytes, including the NULL terminating character
     @param rwflag indicates whether the callback is used for reading/decryption
                   (rwflag=0) or writing/decryption (rwflag=1)
     @param userdata application specific data
     @return the actual length of the password
   */
  static int pem_passwd_cb(char *buf, int size, int rwflag, void *userdata);

  /// Setup Client Context for Certificate Validation
  /**
     @param client_cert_file Client certificate file
     @param root_ca_file Root certificate file
     @param pem_passphrase Optional passphrase for an encrypted private key
  */
  void setup_client_context(const string &client_cert_file,
			    const string &root_ca_file,
			    const string &pem_passphrase = "");

  /// Set the underlying I/O descriptor as non-blocking
  void set_non_blocking();

  /// Check BIO sockect connection
  void check_conn();

  /// Server address
  string _server;

  /// Connection port
  int _port;

  /// BIO Socket
  BIO *_conn;

  /// SSL Context
  SSL_CTX *_ssl_ctx;

  /// SSL object
  SSL *_ssl;

  /// pem passphrase
  static string _pem_passphrase;

  /// Default read/write timeout in seconds
  static const int TIMEOUT;
};

LIBEPP_NICBR_NS_END
#endif //__TRANSPORTTLS_H__
