/* 
 * Copyright (C) 2006 Registro.br. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 1. Redistribution of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY REGISTRO.BR ``AS IS AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIE OF FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL REGISTRO.BR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
/* $Id: DomParser.cpp 379 2006-02-09 20:44:16Z eduardo $ */

#include "config.h"

#include <map>

#include <xercesc/util/XMLUniDefs.hpp>
#include <xercesc/framework/MemBufInputSource.hpp>
#include <xercesc/framework/Wrapper4InputSource.hpp>

#include "DomParser.H"
#include "DomErrorHandler.H"
#include "XmlException.H"

using std::map;

XERCES_CPP_NAMESPACE_USE

LIBEPP_NICBR_NS_BEGIN

DomParser::DomParser(): _builder(0)
{
  try {
    XMLPlatformUtils::Initialize();
  } catch (const XMLException& to_catch) {
    // throws XmlException
    string message = str_transcode(to_catch.getMessage());
    throw XmlException(XmlException::PARSER_INITIALIZE_EXCEPTION,
		       "Could not initialize the XML parser", message);
  }

  /// Instantiate the DOM parser.
  // Load and save module
  static const XMLCh ch[] = { chLatin_L, chLatin_S, chNull };
  DOMImplementation *impl =
    DOMImplementationRegistry::getDOMImplementation(ch);
  DOMImplementationLS *impls = (DOMImplementationLS*)impl;
  _builder = impls->createDOMBuilder(DOMImplementationLS::MODE_SYNCHRONOUS, 0);

  _builder->setErrorHandler(&_error_handler);
}

DomParser::~DomParser()
{
  if (_builder)
    _builder->release();

  // Call the termination method
  XMLPlatformUtils::Terminate();
}

void DomParser::enable_validation(const string& schemas_dir)
{
  string local_schemas_dir = schemas_dir;
  if (schemas_dir == "")
    local_schemas_dir = SCHEMASDIR;

  _builder->setFeature(XMLUni::fgDOMNamespaces, true);
  _builder->setFeature(XMLUni::fgXercesSchema, true);
  _builder->setFeature(XMLUni::fgXercesSchemaFullChecking, true);
  _builder->setFeature(XMLUni::fgDOMValidation, true);
  
  string schema_location =
    "urn:ietf:params:xml:ns:epp-1.0 " + 
    local_schemas_dir + "/epp-1.0.xsd "
    "urn:ietf:params:xml:ns:eppcom-1.0 " + 
    local_schemas_dir + "/eppcom-1.0.xsd "
    "urn:ietf:params:xml:ns:contact-1.0 " + 
    local_schemas_dir + "/contact-1.0.xsd "
    "urn:ietf:params:xml:ns:domain-1.0 " + 
    local_schemas_dir + "/domain-1.0.xsd "
    "urn:ietf:params:xml:ns:host-1.0 " + 
    local_schemas_dir + "/host-1.0.xsd";

  XMLCh *schema_location_ch = XMLString::transcode(schema_location.c_str());

  _builder->setProperty(XMLUni::fgXercesSchemaExternalSchemaLocation, 
			(void*)schema_location_ch);
  
  XMLString::release(&schema_location_ch);
}

void DomParser::disable_validation()
{
  _builder->setFeature(XMLUni::fgDOMNamespaces, false);
  _builder->setFeature(XMLUni::fgXercesSchema, false);
  _builder->setFeature(XMLUni::fgXercesSchemaFullChecking, false);
  _builder->setFeature(XMLUni::fgDOMValidation, false);
}

void DomParser::parse_command(const string &xml_payload)
{
  parse(xml_payload);  
}

auto_ptr<Response> DomParser::parse_login_rsp(const string &xml_payload)
{
  return parse_response(xml_payload);
}

auto_ptr<Response> DomParser::parse_logout_rsp(const string &xml_payload)
{
  return parse_response(xml_payload);
}

auto_ptr<Response> DomParser::parse_response(const string &xml_payload)
{
  auto_ptr<Response> resp(new Response());

  DOMNode *n = parse(xml_payload);
  
  fill_response(n, resp.get());

  return resp;
}

auto_ptr<Greeting> DomParser::parse_greeting(const string &xml_payload)
{
  auto_ptr<Greeting> greeting(new Greeting());

  DOMNode *n = parse(xml_payload);

  fill_greeting(n, greeting.get());

  return greeting;
}

auto_ptr<ContactCheckRsp> DomParser::parse_contact_check_rsp(
                                                     const string &xml_payload)
{
  auto_ptr<ContactCheckRsp> contact_check_rsp(new ContactCheckRsp());

  DOMNode *n = parse(xml_payload);

  fill_contact_check_rsp(n, contact_check_rsp.get());

  return contact_check_rsp;
}

auto_ptr<ContactCreateRsp> DomParser::parse_contact_create_rsp(
                                                     const string &xml_payload)
{
  auto_ptr<ContactCreateRsp> contact_create_rsp(new ContactCreateRsp());

  DOMNode *n = parse(xml_payload);

  fill_contact_create_rsp(n, contact_create_rsp.get());

  return contact_create_rsp;
}

auto_ptr<ContactInfoRsp> DomParser::parse_contact_info_rsp(
                                                     const string &xml_payload)
{
  auto_ptr<ContactInfoRsp> contact_info_rsp(new ContactInfoRsp());

  DOMNode *n = parse(xml_payload);

  fill_contact_info_rsp(n, contact_info_rsp.get());

  return contact_info_rsp;
}

auto_ptr<Response> DomParser::parse_contact_update_rsp(
						     const string &xml_payload)
{
  return parse_response(xml_payload);
}

auto_ptr<DomainCheckRsp> DomParser::parse_domain_check_rsp(
                                                     const string &xml_payload)
{
  auto_ptr<DomainCheckRsp> domain_check_rsp(new DomainCheckRsp());

  DOMNode *n = parse(xml_payload);

  fill_domain_check_rsp(n, domain_check_rsp.get());

  return domain_check_rsp;
}

auto_ptr<DomainCreateRsp> DomParser::parse_domain_create_rsp(
                                                     const string &xml_payload)
{
  auto_ptr<DomainCreateRsp> domain_create_rsp(new DomainCreateRsp());

  DOMNode *n = parse(xml_payload);

  fill_domain_create_rsp(n, domain_create_rsp.get());

  return domain_create_rsp;
}

auto_ptr<DomainInfoRsp> DomParser::parse_domain_info_rsp(
                                                     const string &xml_payload)
{
  auto_ptr<DomainInfoRsp> domain_info_rsp(new DomainInfoRsp());

  DOMNode *n = parse(xml_payload);

  fill_domain_info_rsp(n, domain_info_rsp.get());

  return domain_info_rsp;
}

auto_ptr<DomainRenewRsp> DomParser::parse_domain_renew_rsp(
						     const string &xml_payload)
{
  auto_ptr<DomainRenewRsp> domain_renew_rsp(new DomainRenewRsp());

  DOMNode *n = parse(xml_payload);

  fill_domain_renew_rsp(n, domain_renew_rsp.get());

  return domain_renew_rsp;
}

auto_ptr<Response> DomParser::parse_domain_update_rsp(
					const string &xml_payload)
{
  return parse_response(xml_payload);
}

auto_ptr<PollRsp> DomParser::parse_poll_rsp(const string &xml_payload)
{
  auto_ptr<PollRsp> poll_rsp(new PollRsp());
  
  DOMNode *n = parse(xml_payload);
  
  fill_poll_rsp(n, poll_rsp.get());
  
  return poll_rsp;  
}

string DomParser::str_transcode(const XMLCh *const to_transcode)
{
  char *str = XMLString::transcode(to_transcode);
  string ret_str;
  if (str == 0) {
    ret_str = "";
  } else {
    ret_str = str;
  }
  XMLString::release(&str);
  return ret_str;
}

DOMNode *DomParser::parse(const string &xml_payload)
{
  char *sysId = "";

  MemBufInputSource *tmp_membuf = 
    new MemBufInputSource((const XMLByte*)xml_payload.c_str(),
			  xml_payload.length(), sysId); 
  Wrapper4InputSource *membuf = new Wrapper4InputSource(tmp_membuf);

  DOMDocument *doc = 0;
  try {
    // Reset document vector pool and release memory back to the system
    _builder->resetDocumentPool();
    doc = _builder->parse(*membuf);
  } catch(const XmlException &e) { 
    // if it is a XmlException, it was thrown by the ErrorHandler
    // just rethrow
    throw;
  } catch (...) {
    throw XmlException(XmlException::UNEXPECTED_PARSER_EXCEPTION,
				    "An unexpected exception occurred in XML "
				    "Parser");
  }
  
  if(doc == 0 || doc->getDocumentElement() == 0) {
    throw XmlException(XmlException::NULL_XML_DOCUMENT,
		       "The XML Parser returned a NULL document");
  }

  return (DOMNode*)doc->getDocumentElement();
}

void DomParser::fill_response(DOMNode *n, Response *response)
{  
  bool look_children = false;
  if (n->getNodeType() == DOMNode::ELEMENT_NODE) {
    string elem_name = str_transcode(n->getNodeName());
    if (elem_name == "epp" || elem_name == "response") {
      look_children = true;
    } else if (elem_name == "result") {
      look_children = true;
      DOMNamedNodeMap *attrs = n->getAttributes();
      if (attrs->getLength() > 0) {
	DOMAttr *attr_node = (DOMAttr*) attrs->item(0);
	if (str_transcode(attr_node->getName()) == "code") {
	  string code = str_transcode(attr_node->getValue());
	  response->set_result_code((Response::ResultCode) atoi(code.c_str()));
	}
      }
    } else if (elem_name == "msg") {
      response->set_result_msg(str_transcode(n->getTextContent()));
    } else if (elem_name == "clTRID") {
      response->set_clTRID(str_transcode(n->getTextContent()));
    } else if (elem_name == "svTRID") {
      response->set_svTRID(str_transcode(n->getTextContent()));
    } else if (elem_name == "trID") {
      look_children = true;
    }
  }
  
  if (look_children) {
    DOMNode *child;
    for (child = n->getFirstChild(); child != 0;
	 child = child->getNextSibling()) {
      fill_response(child, response);
    }    
  }
}

void DomParser::fill_greeting(DOMNode *n, Greeting *greeting)
{
  bool look_children = false;
  if (n->getNodeType() == DOMNode::ELEMENT_NODE) {
    string elem_name = str_transcode(n->getNodeName());
    if (elem_name == "epp" || elem_name == "greeting") {
      look_children = true;
    } else if (elem_name == "svID") {
      greeting->set_svID(str_transcode(n->getTextContent()));
    } else if (elem_name == "svDate") {
      greeting->set_svDate(str_transcode(n->getTextContent()));
    } else if (elem_name == "svcMenu") {
      look_children = true;
    } else if (elem_name == "version") {
      greeting->set_version(str_transcode(n->getTextContent()));
    } else if (elem_name == "lang") {
      greeting->set_lang(str_transcode(n->getTextContent()));
    } else if (elem_name == "objURI") {
      greeting->set_objURI(str_transcode(n->getTextContent()));
    } else if (elem_name == "svcExtension") {
      look_children = true;
    } else if (elem_name == "extURI") {
      greeting->set_extURI(str_transcode(n->getTextContent()));
    } else if (elem_name == "dcp") {
      look_children = true;
    } else if (elem_name == "access") {
      DOMNode *child;
      for (child = n->getFirstChild(); child != 0;
	   child = child->getNextSibling()) {
	if (child->getNodeType() == DOMNode::ELEMENT_NODE) {
	  string e_name = str_transcode(child->getNodeName());
	  if (e_name == "all") {
	    greeting->set_access(Greeting::ALL);
	  } else if (e_name == "none") {
	    greeting->set_access(Greeting::NONE_AC);
	  } else if (e_name == "null") {
	    greeting->set_access(Greeting::NULL_AC);
	  } else if (e_name == "other") {
	    greeting->set_access(Greeting::OTHER_AC);
	  } else if (e_name == "personal") {
	    greeting->set_access(Greeting::PERSONAL);
	  } else if (e_name == "personalAndOther") {
	    greeting->set_access(Greeting::PERSONAL_AND_OTHER);
	  }
	}
      }
    } else if (elem_name == "statement") {
      look_children = true;
    } else if (elem_name == "purpose") {
      DOMNode *child;
      for (child = n->getFirstChild(); child != 0;
	   child = child->getNextSibling()) {
	if (child->getNodeType() == DOMNode::ELEMENT_NODE) {
	  string e_name = str_transcode(child->getNodeName());
	  if (e_name == "admin") {
	    greeting->set_purpose(Greeting::ADMIN);
	  } else if (e_name == "contact") {
	    greeting->set_purpose(Greeting::CONTACT);
	  } else if (e_name == "other") {
	    greeting->set_purpose(Greeting::OTHER_PR);
	  } else if (e_name == "prov") {
	    greeting->set_purpose(Greeting::PROV);
	  }
	}
      }      
    } else if (elem_name == "recipient") {
      DOMNode *child;
      for (child = n->getFirstChild(); child != 0;
	   child = child->getNextSibling()) {
	if (child->getNodeType() == DOMNode::ELEMENT_NODE) {
	  string e_name = str_transcode(child->getNodeName());
	  if (e_name == "other") {
	    greeting->set_recipient(Greeting::OTHER_RC);
	  } else if (e_name == "ours") {
	    greeting->set_recipient(Greeting::OURS);
	    DOMNode *recDesc = child->getFirstChild();
	    if (recDesc && recDesc->getNodeType() == DOMNode::ELEMENT_NODE) {
	      if (str_transcode(recDesc->getNodeName()) == "recDesc") {
		greeting->set_recDesc(str_transcode(recDesc->getTextContent()));
	      }
	    }
	  } else if (e_name == "public") {
	    greeting->set_recipient(Greeting::PUBLIC);
	  } else if (e_name == "same") {
	    greeting->set_recipient(Greeting::SAME);
	  } else if (e_name == "unrelated") {
	    greeting->set_recipient(Greeting::UNRELATED);
	  }
	}
      }
    } else if (elem_name == "retention") {
      DOMNode *child;
      for (child = n->getFirstChild(); child != 0;
	   child = child->getNextSibling()) {
	if (child->getNodeType() == DOMNode::ELEMENT_NODE) {
	  string e_name = str_transcode(child->getNodeName());
	  if (e_name == "business") {
	    greeting->set_retention(Greeting::BUSINESS);
	  } else if (e_name == "indefinite") {
	    greeting->set_retention(Greeting::INDEFINITE);
	  } else if (e_name == "legal") {
	    greeting->set_retention(Greeting::LEGAL);
	  } else if (e_name == "none") {
	    greeting->set_retention(Greeting::NONE_RT);
	  } else if (e_name == "stated") {
	    greeting->set_retention(Greeting::STATED);
	  }
	}
      }      
    } else if (elem_name == "expiry") {
      look_children = true;
    } else if (elem_name == "absolute") {
      greeting->set_expiry(0, str_transcode(n->getTextContent()));
    } else if (elem_name == "relative") {
      greeting->set_expiry(1, str_transcode(n->getTextContent()));
    }
  }
  
  if (look_children) {
    DOMNode *child;
    for (child = n->getFirstChild(); child != 0;
	 child = child->getNextSibling()) {
      fill_greeting(child, greeting);
    }
  }
}

void DomParser::fill_contact_check_rsp(DOMNode *n, ContactCheckRsp *c_check_rsp)
{
  bool look_children = false;
  if (n->getNodeType() == DOMNode::ELEMENT_NODE) {
    string elem_name = str_transcode(n->getNodeName());
    if (elem_name == "epp" || elem_name == "response") {
      look_children = true;
    } else if (elem_name == "result") {
      look_children = true;
      DOMNamedNodeMap *attrs = n->getAttributes();
      if (attrs->getLength() > 0) {
	DOMAttr *attr_node = (DOMAttr*) attrs->item(0);
	if (str_transcode(attr_node->getName()) == "code") {
	  string code = str_transcode(attr_node->getValue());
	  c_check_rsp->set_result_code((Response::ResultCode)
				       atoi(code.c_str()));
	}
      }
    } else if (elem_name == "msg") {
      c_check_rsp->set_result_msg(str_transcode(n->getTextContent()));
    } else if (elem_name == "trID") {
      look_children = true;
    } else if (elem_name == "clTRID") {
      c_check_rsp->set_clTRID(str_transcode(n->getTextContent()));
    } else if (elem_name == "svTRID") {
      c_check_rsp->set_svTRID(str_transcode(n->getTextContent()));
    } else if (elem_name == "resData") {
      look_children = true;
    } else if (elem_name == "contact:chkData") {
      look_children = true;
    } else if (elem_name == "contact:cd") {
      DOMNode *child;
      string id, avail, reason;
      for (child = n->getFirstChild(); child != 0;
	   child = child->getNextSibling()) {
	if (child->getNodeType() != DOMNode::ELEMENT_NODE) {
	  continue;
	}
	string e_name = str_transcode(child->getNodeName());
	if (e_name == "contact:id") {
	  DOMNamedNodeMap *attrs = child->getAttributes();
	  if (attrs->getLength() > 0) {
	    DOMAttr *attr_node = (DOMAttr*) attrs->item(0);
	    if (str_transcode(attr_node->getName()) == "avail") {
	      avail = str_transcode(attr_node->getValue());
	    }
	  }
	  id = str_transcode(child->getTextContent());
	} else if (e_name == "contact:reason") {
	  reason = str_transcode(child->getTextContent());
	}
      }
      c_check_rsp->insert_availability(id, avail, reason);
    }
  }  

  if (look_children) {
    DOMNode *child;
    for (child = n->getFirstChild(); child != 0;
         child = child->getNextSibling()) {
      fill_contact_check_rsp(child, c_check_rsp);
    }
  }
}

void DomParser::fill_contact_create_rsp(DOMNode *n,
					ContactCreateRsp *c_create_rsp)
{
  bool look_children = false;
  if (n->getNodeType() == DOMNode::ELEMENT_NODE) {
    string elem_name = str_transcode(n->getNodeName());
    if (elem_name == "epp" || elem_name == "response") {
      look_children = true;
    } else if (elem_name == "result") {
      look_children = true;
      DOMNamedNodeMap *attrs = n->getAttributes();
      if (attrs->getLength() > 0) {
        DOMAttr *attr_node = (DOMAttr*) attrs->item(0);
        if (str_transcode(attr_node->getName()) == "code") {
          string code = str_transcode(attr_node->getValue());
          c_create_rsp->set_result_code((Response::ResultCode)
					atoi(code.c_str()));
        }
      }
    } else if (elem_name == "msg") {
      c_create_rsp->set_result_msg(str_transcode(n->getTextContent()));
    } else if (elem_name == "trID") {
      look_children = true;
    } else if (elem_name == "clTRID") {
      c_create_rsp->set_clTRID(str_transcode(n->getTextContent()));
    } else if (elem_name == "svTRID") {
      c_create_rsp->set_svTRID(str_transcode(n->getTextContent()));
    } else if (elem_name == "resData") {
      look_children = true;
    } else if (elem_name == "contact:creData") {
      look_children = true;
    } else if (elem_name == "contact:id") {
      c_create_rsp->set_id(str_transcode(n->getTextContent()));
    } else if (elem_name == "contact:crDate") {
      c_create_rsp->set_crDate(str_transcode(n->getTextContent()));
    }
  }

  if (look_children) {
    DOMNode *child;
    for (child = n->getFirstChild(); child != 0;
         child = child->getNextSibling()) {
      fill_contact_create_rsp(child, c_create_rsp);
    }
  }
}

void DomParser::fill_contact_info_rsp(DOMNode *n, ContactInfoRsp *c_info_rsp)
{
  bool look_children = false;
  if (n->getNodeType() == DOMNode::ELEMENT_NODE) {
    string elem_name = str_transcode(n->getNodeName());
    if (elem_name == "epp" || elem_name == "response") {
      look_children = true;
    } else if (elem_name == "result") {
      look_children = true;
      DOMNamedNodeMap *attrs = n->getAttributes();
      if (attrs->getLength() > 0) {
        DOMAttr *attr_node = (DOMAttr*) attrs->item(0);
        if (str_transcode(attr_node->getName()) == "code") {
          string code = str_transcode(attr_node->getValue());
          c_info_rsp->set_result_code((Response::ResultCode)
				      atoi(code.c_str()));
        }
      }
    } else if (elem_name == "msg") {
      c_info_rsp->set_result_msg(str_transcode(n->getTextContent()));
    } else if (elem_name == "trID") {
      look_children = true;
    } else if (elem_name == "clTRID") {
      c_info_rsp->set_clTRID(str_transcode(n->getTextContent()));
    } else if (elem_name == "svTRID") {
      c_info_rsp->set_svTRID(str_transcode(n->getTextContent()));
    } else if (elem_name == "resData") {
      look_children = true;
    } else if (elem_name == "contact:infData") {
      look_children = true;
    } else if (elem_name == "contact:id") {
      c_info_rsp->get_common_data()->set_id(str_transcode(n->getTextContent()));
    } else if (elem_name == "contact:roid") {
      c_info_rsp->set_roid(str_transcode(n->getTextContent()));
    } else if (elem_name == "contact:status") {
      DOMNamedNodeMap *attrs = n->getAttributes();
      for (unsigned int i = 0; i < attrs->getLength(); i++) {
	DOMAttr *attr_node = (DOMAttr*) attrs->item(i);
	if (str_transcode(attr_node->getName()) == "s") {
	  c_info_rsp->insert_status(str_transcode(attr_node->getValue()));
	}
      }
    } else if (elem_name == "contact:postalInfo") {
      struct CommonData::PostalInfo pi;
      fill_postal_info(n, &pi);
      c_info_rsp->get_common_data()->insert_postal_info(pi);
    } else if (elem_name == "contact:voice") {
      DOMNamedNodeMap *attrs = n->getAttributes();
      struct CommonData::Phone phone;
      if (attrs->getLength() > 0) {
	DOMAttr *attr_node = (DOMAttr*) attrs->item(0);
	if (str_transcode(attr_node->getName()) == "x") {
	  phone.ext = str_transcode(attr_node->getValue());
	}
      }
      phone.number = str_transcode(n->getTextContent());
      c_info_rsp->get_common_data()->set_voice(phone);
    } else if (elem_name == "contact:fax") {
      DOMNamedNodeMap *attrs = n->getAttributes();
      struct CommonData::Phone phone;
      if (attrs->getLength() > 0) {
	DOMAttr *attr_node = (DOMAttr*) attrs->item(0);
	if (str_transcode(attr_node->getName()) == "x") {
	  phone.ext = str_transcode(attr_node->getValue());
	}
      }
      phone.number = str_transcode(n->getTextContent());
      c_info_rsp->get_common_data()->set_fax(phone);
    } else if (elem_name == "contact:email") {
      string email = str_transcode(n->getTextContent());
      c_info_rsp->get_common_data()->set_email(email);
    } else if (elem_name == "contact:clID") {
      c_info_rsp->set_clID(str_transcode(n->getTextContent()));
    } else if (elem_name == "contact:crID") {
      c_info_rsp->set_crID(str_transcode(n->getTextContent()));
    } else if (elem_name == "contact:crDate") {
      c_info_rsp->set_crDate(str_transcode(n->getTextContent()));
    } else if (elem_name == "contact:upID") {
      c_info_rsp->set_upID(str_transcode(n->getTextContent()));
    } else if (elem_name == "contact:upDate") {
      c_info_rsp->set_upDate(str_transcode(n->getTextContent()));
    } else if (elem_name == "contact:trDate") {
      c_info_rsp->set_trDate(str_transcode(n->getTextContent()));
    } else if (elem_name == "contact:authInfo") {
      DOMNamedNodeMap *attrs = n->getAttributes();
      struct AuthInfo authInfo;
      if (attrs->getLength() > 0) {
	DOMAttr *attr_node = (DOMAttr*) attrs->item(0);
	if (str_transcode(attr_node->getName()) == "roid") {
	  authInfo.set_roid(str_transcode(attr_node->getValue()));
	}
      }
      DOMNode *child;
      for (child = n->getFirstChild(); child != 0;
	   child = child->getNextSibling()) {
	if (child->getNodeType() != DOMNode::ELEMENT_NODE) {
	  continue;
	}
	string e_name = str_transcode(child->getNodeName());
	if (e_name == "contact:pw") {
	  authInfo.set_pw(str_transcode(child->getTextContent()));
	  break;
	}
      }
      c_info_rsp->set_authInfo(authInfo);
    } else if (elem_name == "contact:disclose") {
      CommonData::Disclose disc;
      DOMNamedNodeMap *attrs = n->getAttributes();
      DOMAttr *attr_node;
      bool has_flag = false;
      if (attrs->getLength() > 0) {
	attr_node = (DOMAttr*) attrs->item(0);
	if (str_transcode(attr_node->getName()) == "flag") {
	  has_flag = true;
	  if (str_transcode(attr_node->getValue()) == "0") 
	    disc.flag = false;
	  else
	    disc.flag = true;
	}
      }
      if (has_flag) {
	DOMNode *child;
	for (child = n->getFirstChild(); child != 0;
	     child = child->getNextSibling()) {
	  if (child->getNodeType() != DOMNode::ELEMENT_NODE) {
	    continue;
	  }
	  string e_name = str_transcode(child->getNodeName());
	  if (e_name == "contact:name") {
	    attrs = n->getAttributes();
	    if (attrs->getLength() > 0) {
	      attr_node = (DOMAttr*) attrs->item(0);
	      if (str_transcode(attr_node->getName()) == "loc") {
		disc.name_loc = true;
	      } else if (str_transcode(attr_node->getName()) == "int") {
		disc.name_int = true;
	      }
	    }
	  } else if (e_name == "contact:org") {
	    attrs = n->getAttributes();
	    if (attrs->getLength() > 0) {
	      attr_node = (DOMAttr*) attrs->item(0);
	      if (str_transcode(attr_node->getName()) == "loc") {
		disc.org_loc = true;
	      } else if (str_transcode(attr_node->getName()) == "int") {
		disc.org_int = true;
	      }
	    }
	  } else if (e_name == "contact:addr") {
	    attrs = n->getAttributes();
	    if (attrs->getLength() > 0) {
	      attr_node = (DOMAttr*) attrs->item(0);
	      if (str_transcode(attr_node->getName()) == "loc") {
		disc.addr_loc = true;
	      } else if (str_transcode(attr_node->getName()) == "int") {
		disc.addr_int = true;
	      }
	    }
	  } else if (e_name == "contact:voice") {
	    disc.voice = true;
	  } else if (e_name == "contact:fax") {
	    disc.fax = true;
	  } else if (e_name == "contact:email") {
	    disc.email = true;
	  }
	}
	c_info_rsp->get_common_data()->set_disclose(disc);
      }
    }
  }

  if (look_children) {
    DOMNode *child;
    for (child = n->getFirstChild(); child != 0;
         child = child->getNextSibling()) {
      fill_contact_info_rsp(child, c_info_rsp);
    }
  }
}

void DomParser::fill_domain_check_rsp(DOMNode *n, DomainCheckRsp *d_check_rsp)
{
  bool look_children = false;
  if (n->getNodeType() == DOMNode::ELEMENT_NODE) {
    string elem_name = str_transcode(n->getNodeName());
    if (elem_name == "epp" || elem_name == "response") {
      look_children = true;
    } else if (elem_name == "result") {
      look_children = true;
      DOMNamedNodeMap *attrs = n->getAttributes();
      if (attrs->getLength() > 0) {
        DOMAttr *attr_node = (DOMAttr*) attrs->item(0);
        if (str_transcode(attr_node->getName()) == "code") {
          string code = str_transcode(attr_node->getValue());
          d_check_rsp->set_result_code((Response::ResultCode)
				       atoi(code.c_str()));
        }
      }
    } else if (elem_name == "msg") {
      d_check_rsp->set_result_msg(str_transcode(n->getTextContent()));
    } else if (elem_name == "trID") {
      look_children = true;
    } else if (elem_name == "clTRID") {
      d_check_rsp->set_clTRID(str_transcode(n->getTextContent()));
    } else if (elem_name == "svTRID") {
      d_check_rsp->set_svTRID(str_transcode(n->getTextContent()));
    } else if (elem_name == "resData") {
      look_children = true;
    } else if (elem_name == "domain:chkData") {
      look_children = true;
    } else if (elem_name == "domain:cd") {
      DOMNode *child;
      string name, avail, reason;
      for (child = n->getFirstChild(); child != 0;
	   child = child->getNextSibling()) {
	if (child->getNodeType() != DOMNode::ELEMENT_NODE) {
	  continue;
	}
	string e_name = str_transcode(child->getNodeName());
	if (e_name == "domain:name") {
	  DOMNamedNodeMap *attrs = child->getAttributes();
	  if (attrs->getLength() > 0) {
	    DOMAttr *attr_node = (DOMAttr*) attrs->item(0);
	    if (str_transcode(attr_node->getName()) == "avail") {
	      avail = str_transcode(attr_node->getValue());
	    }
	  }
	  name = str_transcode(child->getTextContent());
	} else if (e_name == "domain:reason") {
	  reason = str_transcode(child->getTextContent());
	}
      }
      d_check_rsp->insert_availability(name, avail, reason);
    }
  }

  if (look_children) {
    DOMNode *child;
    for (child = n->getFirstChild(); child != 0;
         child = child->getNextSibling()) {
      fill_domain_check_rsp(child, d_check_rsp);
    }
  }
}

void DomParser::fill_domain_create_rsp(DOMNode *n,
				       DomainCreateRsp *d_create_rsp)
{
  bool look_children = false;
  if (n->getNodeType() == DOMNode::ELEMENT_NODE) {
    string elem_name = str_transcode(n->getNodeName());
    if (elem_name == "epp" || elem_name == "response") {
      look_children = true;
    } else if (elem_name == "result") {
      look_children = true;
      DOMNamedNodeMap *attrs = n->getAttributes();
      if (attrs->getLength() > 0) {
        DOMAttr *attr_node = (DOMAttr*) attrs->item(0);
        if (str_transcode(attr_node->getName()) == "code") {
          string code = str_transcode(attr_node->getValue());
          d_create_rsp->set_result_code((Response::ResultCode)
					atoi(code.c_str()));
        }
      }
    } else if (elem_name == "msg") {
      d_create_rsp->set_result_msg(str_transcode(n->getTextContent()));
    } else if (elem_name == "trID") {
      look_children = true;
    } else if (elem_name == "clTRID") {
      d_create_rsp->set_clTRID(str_transcode(n->getTextContent()));
    } else if (elem_name == "svTRID") {
      d_create_rsp->set_svTRID(str_transcode(n->getTextContent()));
    } else if (elem_name == "resData") {
      look_children = true;
    } else if (elem_name == "domain:creData") {
      look_children = true;
    } else if (elem_name == "domain:name") {
      d_create_rsp->set_name(str_transcode(n->getTextContent()));
    } else if (elem_name == "domain:crDate") {
      d_create_rsp->set_crDate(str_transcode(n->getTextContent()));
    } else if (elem_name == "domain:exDate") {
      d_create_rsp->set_exDate(str_transcode(n->getTextContent()));
    }
  }

  if (look_children) {
    DOMNode *child;
    for (child = n->getFirstChild(); child != 0;
         child = child->getNextSibling()) {
      fill_domain_create_rsp(child, d_create_rsp);
    }
  }
}

void DomParser::fill_domain_info_rsp(DOMNode *n, DomainInfoRsp *d_info_rsp)
{
  bool look_children = false;
  if (n->getNodeType() == DOMNode::ELEMENT_NODE) {
    string elem_name = str_transcode(n->getNodeName());
    if (elem_name == "epp" || elem_name == "response") {
      look_children = true;
    } else if (elem_name == "result") {
      look_children = true;
      DOMNamedNodeMap *attrs = n->getAttributes();
      if (attrs->getLength() > 0) {
        DOMAttr *attr_node = (DOMAttr*) attrs->item(0);
        if (str_transcode(attr_node->getName()) == "code") {
          string code = str_transcode(attr_node->getValue());
          d_info_rsp->set_result_code((Response::ResultCode)
				      atoi(code.c_str()));
        }
      }
    } else if (elem_name == "msg") {
      d_info_rsp->set_result_msg(str_transcode(n->getTextContent()));
    } else if (elem_name == "trID") {
      look_children = true;
    } else if (elem_name == "clTRID") {
      d_info_rsp->set_clTRID(str_transcode(n->getTextContent()));
    } else if (elem_name == "svTRID") {
      d_info_rsp->set_svTRID(str_transcode(n->getTextContent()));
    } else if (elem_name == "resData") {
      look_children = true;
    } else if (elem_name == "domain:infData") {
      look_children = true;
    } else if (elem_name == "domain:name") {
      d_info_rsp->set_name(str_transcode(n->getTextContent()));
    } else if (elem_name == "domain:roid") {
      d_info_rsp->set_roid(str_transcode(n->getTextContent()));
    } else if (elem_name == "domain:status") {
      DOMNamedNodeMap *attrs = n->getAttributes();
      for (unsigned int i = 0; i < attrs->getLength(); i++) {
	DOMAttr *attr_node = (DOMAttr*) attrs->item(i);
	if (str_transcode(attr_node->getName()) == "s") {
	  d_info_rsp->insert_status(str_transcode(attr_node->getValue()));
	}
      }
    } else if (elem_name == "domain:registrant") {
      d_info_rsp->set_registrant(str_transcode(n->getTextContent()));
    } else if (elem_name == "domain:contact") {
      DOMNamedNodeMap *attrs = n->getAttributes();
      string type;
      for (unsigned int i = 0; i < attrs->getLength(); i++) {
	DOMAttr *attr_node = (DOMAttr*) attrs->item(i);
	if (str_transcode(attr_node->getName()) == "type") {
	  type = str_transcode(attr_node->getValue());
	}
      }
      d_info_rsp->insert_contact(type, str_transcode(n->getTextContent()));
    } else if (elem_name == "domain:ns") {
      look_children = true;
    } else if (elem_name == "domain:hostAttr") {
      NameServer ns;
      DOMNode *child;
      for (child = n->getFirstChild(); child != 0;
	   child = child->getNextSibling()) {
	if (child->getNodeType() != DOMNode::ELEMENT_NODE) {
	  continue;
	}
	string e_name = str_transcode(child->getNodeName());
	if (e_name == "domain:hostName") {
	  ns.name = str_transcode(child->getTextContent());
	} else if (e_name == "domain:hostAddr") {
	  ns.ips.insert(str_transcode(child->getTextContent()));
	}
      }
      d_info_rsp->insert_nameserver(ns);
    } else if (elem_name == "domain:clID") {
      d_info_rsp->set_clID(str_transcode(n->getTextContent()));
    } else if (elem_name == "domain:crID") {
      d_info_rsp->set_crID(str_transcode(n->getTextContent()));
    } else if (elem_name == "domain:crDate") {
      d_info_rsp->set_crDate(str_transcode(n->getTextContent()));
    } else if (elem_name == "domain:upID") {
      d_info_rsp->set_upID(str_transcode(n->getTextContent()));
    } else if (elem_name == "domain:upDate") {
      d_info_rsp->set_upDate(str_transcode(n->getTextContent()));
    } else if (elem_name == "domain:exDate") {
      d_info_rsp->set_exDate(str_transcode(n->getTextContent()));
    } else if (elem_name == "domain:trDate") {
      d_info_rsp->set_trDate(str_transcode(n->getTextContent()));
    } else if (elem_name == "domain:authInfo") {
      DOMNamedNodeMap *attrs = n->getAttributes();
      string roid = "";
      if (attrs->getLength() > 0) {
	DOMAttr *attr_node = (DOMAttr*) attrs->item(0);
	if (str_transcode(attr_node->getName()) == "roid") {
	  roid = str_transcode(attr_node->getValue());
	}
      }
      AuthInfo authInfo;
      authInfo.set_roid(roid);
      authInfo.set_pw(str_transcode(n->getTextContent()));
      d_info_rsp->set_authInfo(authInfo);
    } else if (elem_name == "domain:host") {
      // Ignored
    } else if (elem_name == "domain:hostObj") {
      // Ignored
    }
  }

  if (look_children) {
    DOMNode *child;
    for (child = n->getFirstChild(); child != 0;
         child = child->getNextSibling()) {
      fill_domain_info_rsp(child, d_info_rsp);
    }
  }
}

void DomParser::fill_domain_renew_rsp(DOMNode *n,
				      DomainRenewRsp *d_renew_rsp) 
{
  bool look_children = false;
  if (n->getNodeType() == DOMNode::ELEMENT_NODE) {
    string elem_name = str_transcode(n->getNodeName());
    if (elem_name == "epp" || elem_name == "response") {
      look_children = true;
    } else if (elem_name == "result") {
      look_children = true;
      DOMNamedNodeMap *attrs = n->getAttributes();
      if (attrs->getLength() > 0) {
        DOMAttr *attr_node = (DOMAttr*) attrs->item(0);
        if (str_transcode(attr_node->getName()) == "code") {
          string code = str_transcode(attr_node->getValue());
          d_renew_rsp->set_result_code((Response::ResultCode)
				       atoi(code.c_str()));
        }
      }
    } else if (elem_name == "msg") {
      d_renew_rsp->set_result_msg(str_transcode(n->getTextContent()));
    } else if (elem_name == "trID") {
      look_children = true;
    } else if (elem_name == "clTRID") {
      d_renew_rsp->set_clTRID(str_transcode(n->getTextContent()));
    } else if (elem_name == "svTRID") {
      d_renew_rsp->set_svTRID(str_transcode(n->getTextContent()));
    } else if (elem_name == "resData") {
      look_children = true;
    } else if (elem_name == "domain:renData") {
      look_children = true;
    } else if (elem_name == "domain:name") {
      d_renew_rsp->set_name(str_transcode(n->getTextContent()));
    } else if (elem_name == "domain:exDate") {
      d_renew_rsp->set_exDate(str_transcode(n->getTextContent()));
    }      
  }      
    
  if (look_children) {
    DOMNode *child;
    for (child = n->getFirstChild(); child != 0;
	 child = child->getNextSibling()) {
      fill_domain_renew_rsp(child, d_renew_rsp);
    }
  }
}

void DomParser::fill_poll_rsp(DOMNode *n, PollRsp *poll_rsp)
{
  bool look_children = false;
  if (n->getNodeType() == DOMNode::ELEMENT_NODE) {
    string elem_name = str_transcode(n->getNodeName());
    if (elem_name == "epp" || elem_name == "response") {
      look_children = true;
    } else if (elem_name == "result") {
      // Get the message text from the child element
      DOMNode *child;
      for (child = n->getFirstChild(); child != 0;
	   child = child->getNextSibling()) {
	if (child->getNodeType() == DOMNode::ELEMENT_NODE) {
	  if (str_transcode(child->getNodeName()) == "msg") {
	    poll_rsp->set_result_msg(str_transcode(child->getTextContent()));
	  }
	}
      }
      
      // Get the attribute code
      DOMNamedNodeMap *attrs = n->getAttributes();
      if (attrs->getLength() > 0) {
	DOMAttr *attr_node = (DOMAttr*) attrs->item(0);
	if (str_transcode(attr_node->getName()) == "code") {
	  string code = str_transcode(attr_node->getValue());
	  poll_rsp->set_result_code((Response::ResultCode)
				    atoi(code.c_str()));
	}
      }
      
    } else if (elem_name == "trID") {
      look_children = true;
    } else if (elem_name == "clTRID") {
      poll_rsp->set_clTRID(str_transcode(n->getTextContent()));
    } else if (elem_name == "svTRID") {
      poll_rsp->set_svTRID(str_transcode(n->getTextContent()));
    } else if (elem_name == "msgQ") {
      look_children = true;
      DOMNamedNodeMap *attrs = n->getAttributes();
      for (unsigned int i = 0; i < attrs->getLength(); i++) {
	DOMAttr *attr_node = (DOMAttr*) attrs->item(i);
	if (str_transcode(attr_node->getName()) == "count") {
	  string msg_count = str_transcode(attr_node->getValue());
	  poll_rsp->set_count(msg_count);
	} else if (str_transcode(attr_node->getName()) == "id") {
	  string msg_id = str_transcode(attr_node->getValue());
	  poll_rsp->set_id(msg_id);
	}
      }
    } else if (elem_name == "qDate") {
      poll_rsp->set_qDate(str_transcode(n->getTextContent()));
    } else if (elem_name == "msg") {

      // Get the attribute lang
      DOMNamedNodeMap *attrs = n->getAttributes();
      for (unsigned int i = 0; i < attrs->getLength(); i++) {
	DOMAttr *attr_node = (DOMAttr*) attrs->item(i);
	if (str_transcode(attr_node->getName()) == "lang") {
	  string msg_lang = str_transcode(attr_node->getValue());
	  poll_rsp->set_lang(msg_lang);
	}
      }

      // Run thru the child elements, getting the attributes and 
      // text content
      DOMNode *child;
      int child_count = 0;
      map<string, PollRsp::MsgContent, less<string> > content;
      for (child = n->getFirstChild(); child != 0;
	   child = child->getNextSibling()) {
	if (child->getNodeType() == DOMNode::ELEMENT_NODE) {
	  child_count++;
	  PollRsp::MsgContent msg_content;
	  map< string, string, less<string> > attributes;
	  string child_name = str_transcode(child->getNodeName());
	  DOMNamedNodeMap *child_attrs = child->getAttributes();
	  for (unsigned int i = 0; i < child_attrs->getLength(); i++) {
	    DOMAttr *child_attr_node = (DOMAttr*) child_attrs->item(i);
	    string attr_name = str_transcode(child_attr_node->getName());
	    string attr_value = str_transcode(child_attr_node->getValue());
	    attributes[attr_name] = attr_value;
	  }
	  string child_text_content = str_transcode(child->getTextContent());
	  msg_content.value = child_text_content;
	  msg_content.attributes = attributes;
	  content[child_name] = msg_content;
	}
      }
      poll_rsp->set_content(content);
      
      // If there were no child elements, get text content
      if (child_count == 0) {
	string msg_text_content = str_transcode(n->getTextContent());
	if (msg_text_content != "") {
	  poll_rsp->set_text(msg_text_content);
	}
      }
      
    } else if (elem_name == "resData") {
      look_children = true;
    } else if (elem_name == "contact:chkData") {
      auto_ptr<ContactCheckRsp> response(new ContactCheckRsp());
      fill_contact_check_rsp(n, response.get());
      poll_rsp->set_response((auto_ptr<Response>)response,
			     PollRsp::CONTACT_CHECK);
    } else if (elem_name == "contact:creData") {
      auto_ptr<ContactCreateRsp> response(new ContactCreateRsp());
      fill_contact_create_rsp(n, response.get());
      poll_rsp->set_response((auto_ptr<Response>)response,
			     PollRsp::CONTACT_CREATE);
    } else if (elem_name == "contact:infData") {
      auto_ptr<ContactInfoRsp> response(new ContactInfoRsp());
      fill_contact_info_rsp(n, response.get());
      poll_rsp->set_response((auto_ptr<Response>)response,
			     PollRsp::CONTACT_INFO);      
    } else if (elem_name == "domain:chkData") {
      auto_ptr<DomainCheckRsp> response(new DomainCheckRsp());
      fill_domain_check_rsp(n, response.get());
      poll_rsp->set_response((auto_ptr<Response>)response,
			     PollRsp::DOMAIN_CHECK);
    } else if (elem_name == "domain:creData") {
      auto_ptr<DomainCreateRsp> response(new DomainCreateRsp());
      fill_domain_create_rsp(n, response.get());
      poll_rsp->set_response((auto_ptr<Response>)response,
			     PollRsp::DOMAIN_CREATE);
    } else if (elem_name == "domain:infData") {
      auto_ptr<DomainInfoRsp> response(new DomainInfoRsp());
      fill_domain_info_rsp(n, response.get());
      poll_rsp->set_response((auto_ptr<Response>)response,
			     PollRsp::DOMAIN_INFO);
    } else if (elem_name == "domain:renData") {
      auto_ptr<DomainRenewRsp> response(new DomainRenewRsp());
      fill_domain_renew_rsp(n, response.get());
      poll_rsp->set_response((auto_ptr<Response>)response,
			     PollRsp::DOMAIN_RENEW);
    }
  }
  
  if (look_children) {
    DOMNode *child;
    for (child = n->getFirstChild(); child != 0;
	 child = child->getNextSibling()) {
      fill_poll_rsp(child, poll_rsp);
    }
  }
}

void DomParser::fill_postal_info(DOMNode *n,
				 struct CommonData::PostalInfo *postal_info)
{
  bool look_children = false;
  if (str_transcode(n->getNodeName()) == "contact:postalInfo") {
    look_children = true;
    DOMNamedNodeMap *attrs = n->getAttributes();
    for (unsigned int i = 0; i < attrs->getLength(); i++) {
      DOMAttr *attr_node = (DOMAttr*) attrs->item(i);
      if (str_transcode(attr_node->getName()) == "type") {
	postal_info->type = str_transcode(attr_node->getValue());
	break;
      }
    }
  }

  if (n->getNodeType() == DOMNode::ELEMENT_NODE) {
    string elem_name = str_transcode(n->getNodeName());
    if (elem_name == "contact:name") {
      postal_info->name = str_transcode(n->getTextContent());
    } else if (elem_name == "contact:org") {
      postal_info->org = str_transcode(n->getTextContent());
    } else if (elem_name == "contact:addr") {
      look_children = true;
    } else if (elem_name == "contact:street") {
      if (postal_info->str1 == "") {
	postal_info->str1 = str_transcode(n->getTextContent());
      } else if (postal_info->str2 == "") {
	postal_info->str2 = str_transcode(n->getTextContent());
      } else if (postal_info->str3 == "") {
	postal_info->str3 = str_transcode(n->getTextContent());
      } 
    } else if (elem_name == "contact:city") {
      postal_info->city = str_transcode(n->getTextContent());
    } else if (elem_name == "contact:sp") {
      postal_info->sp = str_transcode(n->getTextContent());
    } else if (elem_name == "contact:pc") {
      postal_info->pc = str_transcode(n->getTextContent());
    } else if (elem_name == "contact:cc") {
      postal_info->cc = str_transcode(n->getTextContent());
    }
  }

  if (look_children) {
    DOMNode *child;
    for (child = n->getFirstChild(); child != 0;
	 child = child->getNextSibling()) {
      fill_postal_info(child, postal_info);
    }
  }
}

LIBEPP_NICBR_NS_END
