/* 
 * Copyright (C) 2006 Registro.br. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 1. Redistribution of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY REGISTRO.BR ``AS IS AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIE OF FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL REGISTRO.BR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
/* $Id: PollTest.cpp 365 2006-02-09 19:18:34Z milena $ */
#include <memory>

#include "libepp_nicbr.H"

#include "PollTest.H"
#include "Poll.H"
#include "FileUtil.H"
#include "IoException.H"
#include "XmlException.H"

using std::auto_ptr;

LIBEPP_NICBR_NS_USE

CPPUNIT_TEST_SUITE_REGISTRATION(PollTest);

PollTest::PollTest() {}

PollTest::~PollTest() {}

void PollTest::setUp() {}

void PollTest::tearDown() {}

void PollTest::set_xml_template_test()
{
  // tests request command
  string to_be_parsed =
     "<command>"
       "<poll op='$(op)$'$(msgID)$/>"
       "<clTRID>$(clTRID)$</clTRID>"
     "</command>";

  Poll poll;
  PollCmd* poll_cmd = poll.get_command();
  poll_cmd->set_op("req");
  poll_cmd->set_msgID("12345"); 
  
  poll.set_xml_template(to_be_parsed, "ABC-12345");
  
  string expected =
    "<command>"
      "<poll op='req'/>"
      "<clTRID>ABC-12345</clTRID>"
     "</command>";

  // tests acknowledgement command
  to_be_parsed =
     "<command>"
       "<poll op='$(op)$'$(msgID)$/>"
       "<clTRID>$(clTRID)$</clTRID>"
     "</command>";

  poll_cmd->reset();
  poll_cmd->set_op("ack");
  poll_cmd->set_msgID("12345");
  
  poll.set_xml_template(to_be_parsed, "ABC-12345");
  
  expected =
    "<command>"
      "<poll op='ack' msgID='12345'/>"
      "<clTRID>ABC-12345</clTRID>"
     "</command>";

  CPPUNIT_ASSERT(expected == poll.get_xml_command());
  
  // tests reset method
  CPPUNIT_ASSERT(poll_cmd->get_op() != "");
  CPPUNIT_ASSERT(poll_cmd->get_msgID() != "");
  poll_cmd->reset();
  CPPUNIT_ASSERT(poll_cmd->get_op() == "");
  CPPUNIT_ASSERT(poll_cmd->get_msgID() == "");
}

void PollTest::command_test()
{
  Poll poll;
  PollCmd *poll_cmd = poll.get_command();

  poll_cmd->set_op("req");
  poll_cmd->set_msgID("12345"); 

  bool exception_caught = false;
  try {
    string xml_template = 
      FileUtil::read_file("../docs/templates/poll.xml");
    poll.set_xml_template(xml_template, "ABC-12345");

    DomParser parser;
    parser.enable_validation("../docs/schemas");
    parser.parse_command(poll.get_xml_command());
  } catch (const IoException &e) {
    exception_caught = true;
    printf("\nIO Exception: code [%d] message [%s]",
	e.get_code(), e.get_msg().c_str());
  } catch (const XmlException &e) {
    exception_caught = true;
    printf("\nXml Exception: code [%d] message [%s] low level message [%s]\n",
	   e.get_code(), e.get_msg().c_str(), e.get_low_level_msg().c_str());
  }

  CPPUNIT_ASSERT(!exception_caught);
}

void PollTest::response_test()
{
  string expected =
    "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>"
    "<epp xmlns=\"urn:ietf:params:xml:ns:epp-1.0\" "
    "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" "
    "xsi:schemaLocation=\"urn:ietf:params:xml:ns:epp-1.0 "
    "epp-1.0.xsd\">"
    "<response>"
    "<result code=\"1301\">"
    "<msg>Command completed successfully; ack to dequeue</msg>"
    "</result>"
    "<msgQ count=\"5\" id=\"12345\">"
    "<qDate>2000-06-08T22:10:00.0Z</qDate>"
    "<msg lang=\"en\">"
    "<bal type=\"currency\">5</bal>"
    "<limit type=\"currency\">100</limit>"
    "<text>Credit Balance Low</text>"
    "</msg>"
    "</msgQ>"
    "<resData>"
    "<contact:chkData "
    "xmlns:contact=\"urn:ietf:params:xml:ns:contact-1.0\" "
    "xsi:schemaLocation=\"urn:ietf:params:xml:ns:contact-1.0 "
    "contact-1.0.xsd\">"
    "<contact:cd>"
    "<contact:id avail=\"1\">btw214</contact:id>"
    "</contact:cd>"
    "<contact:cd>"
    "<contact:id avail=\"0\">mrf348</contact:id>"
    "<contact:reason>In use</contact:reason>"
    "</contact:cd>"
    "<contact:cd>"
    "<contact:id avail=\"1\">sht145</contact:id>"
    "</contact:cd>"
    "</contact:chkData>"
    "</resData>"
    "<trID>"
    "<clTRID>ABC-12345</clTRID>"
    "<svTRID>DEF-54321</svTRID>"
    "</trID>"
    "</response>"
    "</epp>";

  DomParser parser;
  parser.enable_validation("../docs/schemas");
  Poll poll;
  CPPUNIT_ASSERT_NO_THROW(poll.set_response(expected, &parser));

  string result_code = "1301";
  PollRsp* poll_rsp = poll.get_response();
 
  map< string, PollRsp::MsgContent, less<string> > content =
    poll_rsp->get_content();
  map< string, PollRsp::MsgContent, less<string> >::const_iterator it;
 
  CPPUNIT_ASSERT(poll_rsp->get_result_code() == Response::OK_ACK_DEQUEUE);
  
  string response =
    "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>"
    "<epp xmlns=\"urn:ietf:params:xml:ns:epp-1.0\" "
    "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" "
    "xsi:schemaLocation=\"urn:ietf:params:xml:ns:epp-1.0 "
    "epp-1.0.xsd\">"
    "<response>"
    "<result code=\"" + result_code + "\">"
    "<msg>" + poll_rsp->get_result_msg() + "</msg>"
    "</result>"
    "<msgQ count=\"" + poll_rsp->get_count() + "\" id=\"" +
    poll_rsp->get_id() + "\">"
    "<qDate>" + poll_rsp->get_qDate() + "</qDate>"
    "<msg lang=\"" + poll_rsp->get_lang() + "\">" + poll_rsp->get_text();
  
  for (it = content.begin(); it != content.end(); it++) {
    string elem_name = (*it).first;
    PollRsp::MsgContent msg_content = (*it).second;
    map< string, string, less<string> > attrs = msg_content.attributes;
    map< string, string, less<string> >::const_iterator attrs_it;
    response += "<" + elem_name;
    for (attrs_it = attrs.begin(); attrs_it != attrs.end(); attrs_it++) {
      string attr_name = (*attrs_it).first;
      string attr_value = (*attrs_it).second;
      response += " " + attr_name + "=\"" + attr_value + "\"";
    }
    response += ">" + msg_content.value + "</" + elem_name + ">";
  }
  
  response +=
    "</msg>"
    "</msgQ>"
    "<resData>"
    "<contact:chkData "
    "xmlns:contact=\"urn:ietf:params:xml:ns:contact-1.0\" "
    "xsi:schemaLocation=\"urn:ietf:params:xml:ns:contact-1.0 "
    "contact-1.0.xsd\">";
  
  ContactCheckRsp *contact_check_rsp =
    (ContactCheckRsp *)poll_rsp->get_response();
  map< string, ContactCheckRsp::Availability, less<string> > avail = 
    contact_check_rsp->get_availability_list();
  map< string, ContactCheckRsp::Availability, less<string> >::const_iterator avail_it;

  for (avail_it = avail.begin(); avail_it != avail.end(); avail_it++) {
    response += "<contact:cd><contact:id avail=\"" +
      (*avail_it).second.available + "\">"
      + (*avail_it).first + "</contact:id>";
    if (strcmp((*avail_it).second.reason.c_str(), "") != 0) {
      response += "<contact:reason>" + (*avail_it).second.reason + 
	"</contact:reason>";
    }
    response += "</contact:cd>";
  }
  response += 
    "</contact:chkData>"
    "</resData>"
    "<trID>"
    "<clTRID>" + poll_rsp->get_clTRID() + "</clTRID>"
    "<svTRID>" + poll_rsp->get_svTRID() + "</svTRID>"
    "</trID>"
    "</response>"
    "</epp>";
  
  CPPUNIT_ASSERT_EQUAL(expected, response);
}
